@@ -2,25 +2,12 @@
 
 // Copyright 2021 Oxide Computer Company
 
-// NOTE: Although this file compiles and is semi-useful on macOS, it can't be used to generate
-// probes that DTrace actually understands. In particular, running `dtrace -l` will list the
-// probes, but doing anything on the basis of a probe _firing_ is meaningless.
-//
-// This is because Apple seems to use a different definition for the base address of a probe
-// function in the kernel than other systems. See:
-// https://github.com/apple/darwin-xnu/blob/8f02f2a044b9bb1ad951987ef5bab20ec9486310/bsd/dev/dtrace/dtrace.c#L9394
-// for a note indicating the platform difference. It's not clear why they do this, but it does mean
-// that the probes are not actually usable when running with the no-linker feature flag. This flag
-// is mostly used for testing the code to the extent possible on any system.
-
-use std::{convert::TryFrom, ffi::CString};
-
+use crate::record::{process_section, PROBE_REC_VERSION};
+use crate::{common, module_ident_for_provider, DataType, Probe, Provider};
 use dof::{serialize_section, Section};
 use proc_macro2::TokenStream;
-use quote::{format_ident, quote};
-
-use crate::record::{process_section, PROBE_REC_VERSION};
-use crate::{common, DataType};
+use quote::quote;
+use std::convert::TryFrom;
 
 /// Compile a DTrace provider definition into Rust tokens that implement its probes.
 pub fn compile_provider_source(
@@ -31,7 +18,17 @@ pub fn compile_provider_source(
     let providers = dfile
         .providers()
         .iter()
-        .map(|provider| compile_provider(provider, &config))
+        .map(|provider| {
+            let provider = Provider::from(provider);
+            let tokens = compile_provider(&provider, &config);
+            let mod_name = module_ident_for_provider(&provider);
+            quote! {
+                #[macro_use]
+                pub(crate) mod #mod_name {
+                    #tokens
+                }
+            }
+        })
         .collect::<Vec<_>>();
     Ok(quote! {
         #(#providers)*
@@ -39,21 +36,18 @@ pub fn compile_provider_source(
 }
 
 pub fn compile_provider_from_definition(
-    provider: &dtrace_parser::Provider,
+    provider: &Provider,
     config: &crate::CompileProvidersConfig,
 ) -> TokenStream {
     compile_provider(provider, config)
 }
 
-fn compile_provider(
-    provider: &dtrace_parser::Provider,
-    config: &crate::CompileProvidersConfig,
-) -> TokenStream {
-    let mod_name = format_ident!("__usdt_private_{}", provider.name());
+fn compile_provider(provider: &Provider, config: &crate::CompileProvidersConfig) -> TokenStream {
+    let mod_name = module_ident_for_provider(&provider);
     let probe_impls = provider
-        .probes()
+        .probes
         .iter()
-        .map(|probe| compile_probe(probe, provider.name(), config))
+        .map(|probe| compile_probe(provider, probe, config))
         .collect::<Vec<_>>();
     quote! {
         #[macro_use]
@@ -64,43 +58,44 @@ fn compile_provider(
 }
 
 fn compile_probe(
-    probe: &dtrace_parser::Probe,
-    provider_name: &str,
+    provider: &Provider,
+    probe: &Probe,
     config: &crate::CompileProvidersConfig,
 ) -> TokenStream {
-    let (unpacked_args, in_regs) = common::construct_probe_args(probe.types());
-    let is_enabled_rec = asm_rec(provider_name, probe.name(), None);
-    let probe_rec = asm_rec(provider_name, probe.name(), Some(probe.types()));
+    let (unpacked_args, in_regs) = common::construct_probe_args(&probe.types);
+    let is_enabled_rec = asm_rec(&provider.name, &probe.name, None);
+    let probe_rec = asm_rec(&provider.name, &probe.name, Some(&probe.types));
     let pre_macro_block = TokenStream::new();
     let impl_block = quote! {
-        let mut is_enabled: u64;
-        // TODO can this block be option(pure)?
-        unsafe {
-            asm!(
-                "990:   clr rax",
-                #is_enabled_rec,
-                out("rax") is_enabled,
-                options(nomem, nostack, preserves_flags)
-            );
-        }
-
-        if is_enabled != 0 {
-            #unpacked_args
+        {
+            let mut is_enabled: u64;
             unsafe {
                 asm!(
-                    "990:   nop",
-                    #probe_rec,
-                    #in_regs
+                    "990:   clr rax",
+                    #is_enabled_rec,
+                    out("rax") is_enabled,
                     options(nomem, nostack, preserves_flags)
                 );
             }
+
+            if is_enabled != 0 {
+                #unpacked_args
+                unsafe {
+                    asm!(
+                        "990:   nop",
+                        #probe_rec,
+                        #in_regs
+                        options(nomem, nostack, preserves_flags)
+                    );
+                }
+            }
         }
     };
     common::build_probe_macro(
         config,
-        provider_name,
-        probe.name(),
-        probe.types(),
+        provider,
+        &probe.name,
+        &probe.types,
         pre_macro_block,
         impl_block,
     )
@@ -219,8 +214,10 @@ pub fn register_probes() -> Result<(), crate::Error> {
     }
 }
 
-#[cfg(not(target_os = "macos"))]
 fn ioctl_section(buf: &[u8], modname: [std::os::raw::c_char; 64]) -> Result<(), std::io::Error> {
+    use std::fs::OpenOptions;
+    use std::os::unix::io::AsRawFd;
+
     let helper = dof::dof_bindings::dof_helper {
         dofhp_mod: modname,
         dofhp_addr: buf.as_ptr() as u64,
@@ -228,47 +225,14 @@ fn ioctl_section(buf: &[u8], modname: [std::os::raw::c_char; 64]) -> Result<(),
     };
     let data = &helper as *const _;
     let cmd: i32 = 0x64746803;
-    let ret = unsafe {
-        let file = CString::new("/dev/dtrace/helper".as_bytes()).unwrap();
-        let fd = libc::open(file.as_ptr(), libc::O_RDWR);
-        if fd < 0 {
-            fd
-        } else {
-            libc::ioctl(fd, cmd, data)
-        }
-    };
-    if ret == 0 {
-        Ok(())
-    } else {
+    let file = OpenOptions::new()
+        .read(true)
+        .write(true)
+        .open("/dev/dtrace/helper")?;
+    if unsafe { libc::ioctl(file.as_raw_fd(), cmd, data) } < 0 {
         Err(std::io::Error::last_os_error())
-    }
-}
-
-#[cfg(target_os = "macos")]
-fn ioctl_section(buf: &[u8], modname: [std::os::raw::c_char; 64]) -> Result<(), std::io::Error> {
-    let helper = dof::dof_bindings::dof_ioctl_data {
-        dofiod_count: 1,
-        dofiod_helpers: [dof::dof_bindings::dof_helper {
-            dofhp_mod: modname,
-            dofhp_addr: buf.as_ptr() as u64,
-            dofhp_dof: buf.as_ptr() as u64,
-        }],
-    };
-    let data = &(&helper) as *const _;
-    let cmd: u64 = 0x80086804;
-    let ret = unsafe {
-        let file = CString::new("/dev/dtracehelper".as_bytes()).unwrap();
-        let fd = libc::open(file.as_ptr(), libc::O_RDWR);
-        if fd < 0 {
-            fd
-        } else {
-            libc::ioctl(fd, cmd, data)
-        }
-    };
-    if ret == 0 {
-        Ok(())
     } else {
-        Err(std::io::Error::last_os_error())
+        Ok(())
     }
 }
 
@@ -280,7 +244,10 @@ mod tests {
     fn test_asm_rec() {
         let provider = "provider";
         let probe = "probe";
-        let types = [DataType::U8, DataType::String];
+        let types = [
+            DataType::Native(dtrace_parser::DataType::U8),
+            DataType::Native(dtrace_parser::DataType::String),
+        ];
         let record = asm_rec(provider, probe, Some(&types));
         let mut lines = record.lines();
         println!("{}", record);
